using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using TMPro;
using UnityEngine;
using UnityEngine.Events;


public class GrapplingGunBehaviour : MonoBehaviour, Messages.IUse
{   
    #region objs
    [SkipSerialisation]
    private SpawnableAsset projectileAsset = ModAPI.FindSpawnable("Hook -TGGM");
    [SkipSerialisation]
    public GameObject gunRef;
    [SkipSerialisation]
    public GameObject hookRef;
    [SkipSerialisation]
    public DistanceJoint2D winch;
    [SkipSerialisation]
    private Rigidbody2D rb;
    [SkipSerialisation]
    private PhysicalBehaviour phys;
    #endregion

    #region vectors
    public Vector2 barrelPosition;
    public Vector2 barrelDirection;
    public Vector2 GetBarrelPosition() => (Vector2) this.transform.TransformPoint((Vector3) this.barrelPosition);
    public Vector2 GetBarrelDirection() => (Vector2) (this.transform.TransformDirection((Vector3) this.barrelDirection) * this.transform.localScale.x);
    #endregion

    #region vars
    [SkipSerialisation]
    private readonly HashSet<UnityEngine.Object> linkedObjects = new HashSet<UnityEngine.Object>();
    [SkipSerialisation]
    private bool winchMoving;
    [SkipSerialisation]
    private bool launched;
    [SkipSerialisation]
    private int ammo;
    [SkipSerialisation]
    private float currentDistance;
    [SkipSerialisation]
    public float projectileLaunchStrength = 2.0f;
    [SkipSerialisation]
    public float recoilMultiplier = 0.1f; 
    #endregion

    #region evt
    [SkipSerialisation]
    public UnityEvent OnLaunchEvent;
    [SkipSerialisation]
    public event EventHandler<GameObject> OnLaunch;
    #endregion

    #region setup
    private void Awake()
    {
        this.rb = this.GetComponent<Rigidbody2D>();
        this.phys = this.GetComponent<PhysicalBehaviour>();
        this.gunRef = null;
        this.hookRef = null;
        this.launched = false;
        this.winchMoving = false;
        this.ammo = 1;
    } 
    private void Start(){
        UnityEngine.Object.Destroy(gunRef.GetComponent<FirearmBehaviour>());

        this.winch = gunRef.AddComponent<DistanceJoint2D>();

        this.winch.autoConfigureDistance = false;
        this.winch.maxDistanceOnly = true;
        this.winch.distance = Mathf.Infinity;
        this.winch.anchor = new Vector2(0, 0);
        this.winch.enabled = false;

        this.AddContextMenu();    
    }
    private void AddContextMenu(){

        List<ContextMenuButton> buttons = gunRef.GetComponent<PhysicalBehaviour>().ContextMenuOptions.Buttons;
        ContextMenuButton[] contextMenuButtonArray = new ContextMenuButton[3];

        contextMenuButtonArray[0] = new ContextMenuButton("Stop Winch", "Stop", "Stop the grappling gun's winch", () =>
            {
                if(this.winchMoving != false) 
                {
                    StopWinch();
                }
            });

        contextMenuButtonArray[1] = new ContextMenuButton("Retract Winch", "Retract", "Retract the grappling gun's winch", () =>
            {
                if(!this.winchMoving) 
                {
                    InvokeRepeating("ReelInWinch", 0.2f, 0.1f);
                }else
                {
                    CancelInvoke("ReelInWinch");
                    CancelInvoke("ExtendWinch");
                    InvokeRepeating("ReelInWinch", 0.2f, 0.1f);
                }
            });

        contextMenuButtonArray[2] = new ContextMenuButton("Extend Winch", "Extend", "Extend the grappling gun's winch", () =>
            {
                if(!this.winchMoving) 
                {
                    InvokeRepeating("ExtendWinch", 0.2f, 0.01f);
                }else
                {
                    CancelInvoke("ExtendWinch");
                    CancelInvoke("ReelInWinch");
                    InvokeRepeating("ExtendWinch", 0.2f, 0.01f);
                }
            });

        buttons.AddRange((IEnumerable<ContextMenuButton>) contextMenuButtonArray);
    }
    #endregion

    #region runtime
    private void Update()
    {
        if(this.launched) ModAPI.Draw.Line(this.gunRef.transform.position, this.hookRef.transform.position);
    }
    public void Use(ActivationPropagation activation)
    {
        if (!this.enabled)
            return;
        if(!this.launched)
        {
            this.Launch();
        }else
        {
            this.Recover();
        }
    }
    #endregion

    #region hook
    public void Launch() => this.StartCoroutine(this.LaunchRoutine());
    public void Recover() => this.StartCoroutine(this.RecoverRoutine());
    private IEnumerator LaunchRoutine()
    {
        if(this.ammo == 1)
        {
            ammo--;
            this.launched = true;
            foreach (UnityEngine.Object linkedObject in this.linkedObjects)
            {
                UnityEngine.Object.Destroy(linkedObject);
                UnityEngine.Object.Destroy(this.hookRef);
            }
            this.hookRef = null;

            Quaternion rotation = (double) this.transform.lossyScale.x < 0.0 ? Quaternion.Euler(0.0f, 0.0f, 180f) * this.transform.rotation : this.transform.rotation;
            PhysicalBehaviour component;
            this.hookRef = UnityEngine.Object.Instantiate<GameObject>(this.projectileAsset.Prefab, (Vector3) this.GetBarrelPosition(), rotation, this.gunRef.transform.parent);
            component = hookRef.GetComponent<PhysicalBehaviour>();
            component.SpawnSpawnParticles = false;
            this.hookRef.name = this.projectileAsset.name;
            CatalogBehaviour.PerformMod(this.projectileAsset, hookRef);
            this.hookRef.layer = -1;
            EventHandler<GameObject> onLaunch = this.OnLaunch;
            this.OnLaunchEvent?.Invoke();
            component.rigidbody.AddForce((Vector2) (this.hookRef.transform.right * this.projectileLaunchStrength), ForceMode2D.Impulse);
            this.rb.AddForce((Vector2) (-this.projectileLaunchStrength * this.recoilMultiplier * this.hookRef.transform.right), ForceMode2D.Impulse);
            this.linkedObjects.Add((UnityEngine.Object) hookRef);


            this.hookRef.GetComponent<HookBehaviour>().connectedGun = this;
            this.winch.enabled = true;
            this.winch.connectedBody = hookRef.GetComponent<Rigidbody2D>();
            this.winch.connectedAnchor = new Vector2(0, 0);
            yield return (object) new WaitForFixedUpdate();
        }
    }
    private IEnumerator RecoverRoutine()
    {
        CancelInvoke();
        foreach (UnityEngine.Object linkedObject in this.linkedObjects)
        {
            UnityEngine.Object.Destroy(linkedObject);
            UnityEngine.Object.Destroy(this.hookRef);
        }
        this.hookRef = null;
        this.winch.distance = Mathf.Infinity;
        this.winch.enabled = false;
        this.winchMoving = false;
        this.ammo++;
        this.launched = false; 
        yield return (object) new WaitForFixedUpdate();        
    }
    #endregion
    
    #region Winch
    public void UpdateDistance()
    {
        this.currentDistance = Mathf.Infinity;
        this.currentDistance = Vector2.Distance(this.gunRef.transform.position, this.hookRef.transform.position);
        this.winch.distance = this.currentDistance;
    }
    private void ReelInWinch()
    {
        if(launched == true)
        {
            if(this.winch.distance > 0.5f)
            {
                if(!this.winchMoving) this.winchMoving = true;
                this.winch.distance -= 0.025f;
            }else
            {
                CancelInvoke("ReelInWinch");
                if(this.winchMoving) this.winchMoving = false;
            }
        }
    }
    private void ExtendWinch()
    {
        if(launched == true){
            if(this.winch.distance < currentDistance)
            {
                if(!this.winchMoving) this.winchMoving = true;
                this.winch.distance += 0.025f;
            }else
            {
                CancelInvoke("ExtendWinch");
                if(this.winchMoving) this.winchMoving = false;
            }
        }
    }
    private void  StopWinch()
    {
        CancelInvoke("ExtendWinch");
        CancelInvoke("ReelInWinch");
        if(this.winchMoving) this.winchMoving = false;
    }
    #endregion

    #region Destroy
    private void OnDestroy()
    {
        foreach (UnityEngine.Object linkedObject in this.linkedObjects)
            {
                UnityEngine.Object.Destroy(linkedObject);
                UnityEngine.Object.Destroy(this.hookRef);
            }
        this.hookRef = null;
    } 
    #endregion

}

//------------------------------------------------------------------------------------------------------------------------------------------------------------------------

public class HookBehaviour : MonoBehaviour, Messages.ILodged
{
    #region vars
    [SkipSerialisation]
    public GrapplingGunBehaviour connectedGun;

    [SkipSerialisation]
    private bool lodgedBefore = false;
    #endregion

    #region functions
    public void Lodged(Stabbing stabbing)
    {
        if(!this.lodgedBefore)
        {
            this.lodgedBefore = true;
            this.connectedGun.UpdateDistance();
        }
    }
    #endregion
}